#include "config.h"
#include <stdio.h>
#include <stdlib.h>
#ifdef HAVE_WINDOWS_H
#include <windows.h>
#define FD_SETSIZE 1024
#include <winsock.h>
#define getpid()	_getpid()
#define __ANSI_CPP__
#else
#include <unistd.h>
#include <sys/time.h>
#include <sys/socket.h>
#ifdef HAVE_SYS_SELECT_H
#include <sys/select.h>
#endif
#ifdef HAVE_SYS_UIO_H
#include <sys/uio.h>
#endif
#include <sys/un.h>
#include <netinet/in.h>
#include <netdb.h>
#include <signal.h>
#endif
#include <sys/types.h>
#include <stdio.h>
#include <fcntl.h>
#include <errno.h>
#include <signal.h>
#include <memory.h>

#include "io.h"
#include "io_interface.h"
#include "io_internal.h"
#include "unix_defs.h"
#include <malloc.h>
#include <string.h>
#include <assert.h>

typedef struct format_list {
    IOFormat	format;
    time_t	last_reference;
    struct format_list *next;
} format_list;

typedef struct _format_server {
    char *exchange_name;	/* my exchange_name */
    int exch_pid;		/* my exch_pid */
    int port;			/* server port as announced to world */
    char *hostname;		/* server's hostname */
    void *conn_sock_inet;	/* public sock for IP socket connection */
    char *usock_name;		/* file name for unix socket connection */
    int portCount;		/* number of current ports */
    FSClient *ports;	/* array of ports index by socket fd */
    fd_set fdset;		/* bitmap of those conns */
    char *data_buffer;		/* buffer for data */
    int buffer_size;
    format_list	*lists[2][1];    /* first dimension byte order, second float format */
} _fsserver;

static FSClient
format_server_accept_conn_sock ARGS((format_server fs, void *conn_sock));
static format_server format_server_create();
static void format_server_handle_data ARGS((format_server fs, FSClient fsc));
static int format_server_listen ARGS((format_server fs, int port_num));
static int format_server_poll_and_handle ARGS((format_server fs));

#ifndef HAVE_WINDOWS_H
#ifndef SELECT_DEFINED
extern int select ARGS((int width, fd_set *readfds, fd_set *writefds, 
			fd_set *exceptfds, struct timeval *timeout));
#endif
#ifndef BIND_DEFINED
extern int bind ARGS((int s, struct sockaddr *name, int namelen));
#endif
#ifndef LISTEN_DEFINED
extern int listen ARGS((int s, int backlog));
#endif
#ifndef ACCEPT_DEFINED
extern int accept ARGS((int s, struct sockaddr *addr, int *addrlen));
#endif
extern int socket ARGS((int domain, int type, int protocol));
#ifndef CONNECT_DEFINED
extern int connect ARGS((int s, struct sockaddr *name, int namelen));
#endif
#ifndef SETSOCKOPT_DEFINED
extern int setsockopt ARGS((int s, int level, int optname, const char *optval, int optlen));
#endif
#ifndef GETSOCKNAME_DEFINED
extern int getsockname ARGS((int s, struct sockaddr *name, int *namelen));
#endif
#ifndef GETHOSTNAME_DEFINED
extern int gethostname ARGS((char *name, int namelen));
#endif
#endif
#if defined(HAVE_GETDOMAINNAME) && !defined(GETDOMAINNAME_DEFINED)
extern int getdomainname ARGS((char *name, int namelen));
#endif
#if !defined(GETPEERNAME_DEFINED) & !defined(HAVE_WINDOWS_H)
extern int getpeername ARGS((int, struct sockaddr *, int *));
#endif
extern time_t time();
extern pid_t getpid();

static int no_fork = 0;


static void 
die_with_error(sig)
int sig;
{
    fprintf(stderr, "Format server not responding.  Timeout.\n");
    exit(1);
}

void
main(argc, argv)
int argc; char **argv;
{
    int i;
    format_server fs;
    IOFile test_file;
    char testing_char = 'T';

    for (i=1; i<argc; i++) {
	if (strcmp(argv[i], "-no_fork") == 0) {
	    no_fork++;
	} else {
	    fprintf(stderr, "Unknown argument \"%s\"\n", argv[i]);
	    exit(1);
	}
    }
    
    /* test to see if format server is running */
    test_file = new_IOFile();
#ifndef HAVE_WINDOWS_H
    signal(SIGALRM, die_with_error);
    alarm(30);
#endif
    establish_server_connection(test_file, /* nofallback */ 0);
    if (AtomicWrite(test_file->server_fd, &testing_char, 1, test_file) 
	== 1) {
	/* already running */
	exit(0);
#ifdef HAVE_FORK
    } else {
	/* make really, really certain there's no format_server running */
	FILE *format_server_pid_file = fopen("/tmp/format_server_pid", "r");
	alarm(0);
	if (format_server_pid_file != NULL) {
	    long server_pid;
	    if (fscanf(format_server_pid_file, "%lx", &server_pid) == 1) {
		if (kill(server_pid, 0) == 0 ) {
		    fprintf(stderr, "Format Server %lx not responding, but still running\n",
			   server_pid);
		    exit(0);
		}
	    }
	    fclose(format_server_pid_file);
	}

	if (!no_fork) {
	    if (fork() != 0) {
		/* I'm the parent, return now */
		exit(0);
	    }
	}
	format_server_pid_file = fopen("/tmp/format_server_pid", "w");
	if (format_server_pid_file) {
	    fprintf(format_server_pid_file, "%lx\n", (long)getpid());
	    fclose(format_server_pid_file);
	}
	if (!no_fork) {
	    if (fork() != 0) {
		/* I'm the parent, return now */
		exit(0);
	    }
	}
#endif
    }

    fs = format_server_create();
    if (format_server_listen(fs, FS_PORT) != -1) {
/*	printf("DataExchange server listening at Inet host/port <%s %d>\n",
	       format_server_host_name(fs), format_server_inet_port(fs));*/
	while (1) {
	    format_server_poll_and_handle(fs);
	}
    }
}

static void FSClient_close(FSClient fsc);

static int
format_server_poll_and_handle(format_server fs)
{
    int i, res;
    fd_set rd_set = fs->fdset;

    /* scan existing conn for incoming data & do something */
    res = select(FD_SETSIZE, &rd_set, (fd_set *) NULL,
		 (fd_set *) NULL, NULL);
    if (res == -1) {
	if (errno == EBADF) {
	    int i;
	    int found_one = 0;
	    struct timeval timeout;
	    for (i = 0; i< FD_SETSIZE; i++) {
		if (FD_ISSET(i, &rd_set)) {
		    fd_set test_set;
		    timeout.tv_usec = 0;
		    timeout.tv_sec = 0;
		    FD_ZERO(&test_set);
		    FD_SET(i, &test_set);
		    errno = 0;
		    select(FD_SETSIZE, &test_set, (fd_set*)NULL, 
			   (fd_set *) NULL, &timeout);
		    if (errno == EBADF) {
			fprintf(stderr, "Select failed, fd %d\n is bad.  Removing from select list.\n",
				i);
			FSClient_close(fs->ports[i]);
			found_one++;
		    }
		}
	    }
	    if (found_one == 0) {
		fprintf(stderr, "Bad file descriptor in select.  Failed to localize.  Exiting.\n");
		exit(1);
	    }
	} else if (errno == EINTR) {
	    return 0;
	}
	fprintf(stderr, "select failed, errno %d\n", errno);
	return -1;
    } else if (res == 0)
	return 0;

    if ((long)fs->conn_sock_inet >= 0 &&
	FD_ISSET((unsigned long)fs->conn_sock_inet, &rd_set)) {
	FD_CLR((unsigned long) fs->conn_sock_inet, &rd_set);
	format_server_accept_conn_sock(fs, fs->conn_sock_inet);
    }
    for (i = 0; i < FD_SETSIZE; i++) {	/* check each possible read fd */
	FSClient fsc = fs->ports[i];
	if (fsc == NULL || !FD_ISSET(i, &rd_set))
	    continue;
	format_server_handle_data(fs, fsc);
	
    }
    return res;
}

/*
typedef enum {
    write_record_state, write_format_state, not_blocked
} block_type;
extern unsigned alarm ARGS((unsigned));
static void record_blocking_state ARGS((block_type state,
					FSClient port));
static int block_count = 0;
static FSClient block_port = NULL;
static block_type block_state = not_blocked;

static void
clear_blocking_state()
{
    block_port = NULL;
    block_state = not_blocked;
    block_count++;
}

static void
record_blocking_state(state, port)
block_type state;
FSClient port;
{
    block_port = port;
    block_state = state;
}

static int last_block_count = 0;

static void
check_blocking_state(signal)
int signal;
{
    if ((block_count == last_block_count) && (block_state != not_blocked)
	&& (block_port != NULL)) {
	fprintf(stderr, "Warning!  No forward progress!!\n");
	fprintf(stderr, " Writing to port [%s, %d] \"%s\".\n",
		block_port->hostname, block_port->port,
		block_port->exchange_name);
    }
    last_block_count = block_count;
    alarm(1);
}


void
DEinit_block_check()
{
    struct sigaction sigact;
    sigact.sa_flags = 0;
    sigact.sa_handler = check_blocking_state;
    sigemptyset(&sigact.sa_mask);
    sigaddset(&sigact.sa_mask, SIGALRM);
    sigaction(SIGALRM, &sigact, NULL);
    alarm(1);
}
*/

static int
format_eq(form1, form2)
IOFormat form1;
IOFormat form2;
{
    int i;
    int all_zero = 1;
    int all_eq = 1;
    for (i=0 ; i< (sizeof(form1->server_ID)/sizeof(form1->server_ID[0])); i++) {
	if (form1->server_ID[i] != 0) {
	    all_zero = 0;
	}
	if (form1->server_ID[i] != form2->server_ID[i]) {
	    all_eq = 0;
	}
    }
    if (all_eq && !all_zero) return 1;
    if (form1->field_count != form2->field_count) return 0;
    if (strcmp(form1->format_name, form2->format_name) != 0) return 0;
    if (form1->record_length != form2->record_length) return 0;
    for (i=0; i< form1->field_count; i++) {
	if (strcmp(form1->field_list[i].field_name,
		   form2->field_list[i].field_name) != 0) return 0;
	if (strcmp(form1->field_list[i].field_type,
		   form2->field_list[i].field_type) != 0) return 0;
	if (form1->field_list[i].field_size != 
	    form2->field_list[i].field_size) return 0;
	if (form1->field_list[i].field_offset != 
	    form2->field_list[i].field_offset) return 0;
    }
    return 1;
}

static IOFormat
find_format(fs, fsc, ioformat, new_format_mode)
format_server fs;
FSClient fsc;
IOFormat ioformat;
int new_format_mode;
{
    format_list *list = fs->lists[fsc->byte_reversal][0];
    format_list *last = NULL, *new = NULL;
    /* always, search list for this hosts byte order */
    while (list != NULL) {
	if (format_eq(ioformat, list->format)) {
	    free_IOformat(ioformat);
	    return list->format;
	}
	last = list;
	list = list->next;
    }
    if (new_format_mode) {
	/* didn't find it in the list, create a blank */
	new = (format_list*)io_malloc(sizeof(format_list));
	new->format = ioformat;
	memcpy(&ioformat->server_ID[0], ioformat, sizeof(ioformat));
	memset(&ioformat->server_ID[0] + sizeof(ioformat), 0, 
	       sizeof(ioformat->server_ID[0]) - sizeof(ioformat));
	new->next = NULL;
	if (last == NULL) {
	    fs->lists[fsc->byte_reversal][0] = new;
	} else {
	    last->next = new;
	}
	return ioformat;
    } else {
	/* search other list */
	int other_byte_order = fsc->byte_reversal ? 0 : 1;
	list = fs->lists[other_byte_order][0];
	while (list != NULL) {
	    if (format_eq(ioformat, list->format)) {
		free_IOformat(ioformat);
		return list->format;
	    }
	    last = list;
	    list = list->next;
	}
	return NULL;
    }
}

static void
server_write_id(fsc, ioformat)
FSClient fsc;
IOFormat ioformat;
{
    int junk_errno;
    char *junk_str;
    os_server_write_func(fsc->fd, &ioformat->server_ID[0], 8, &junk_errno,
			 &junk_str);
}

/* 
 * Close an on-line connection.
 */
static void
FSClient_close(FSClient fsc)
{
    format_server fs = fsc->fs;
    int fd = (int) (long) fsc->fd;
    if (fsc->fd != (void*) 0) {
	fs->ports[fd] = NULL;
	FD_CLR((unsigned long)fd, &fs->fdset);
	os_close_func(fsc->fd);
    }
    io_free(fsc);
}

static void
format_server_handle_data(fs, fsc)
format_server fs;
FSClient fsc;
{
    IOFormat ioformat;
    char	next_action;
    if (serverAtomicRead(fsc->fd, &next_action, 1) != 1) {
	FSClient_close(fsc);
	return;
    }
    switch(next_action) {
    case 'T': 
	if (no_fork) {
	    printf("Testing\n");
	}
	break;
    case 'F':
	if (no_fork) {
	    printf("Format Registration\n");
	}
	ioformat = server_read_format(fs, fsc);
	ioformat = find_format(fs, fsc, ioformat,
			       1 /* new format mode */);
	if (no_fork) {
	    printf("Returning -> ");
	    print_format_ID(ioformat);
	    printf("\n");
	    dump_IOFormat(ioformat);
	}
	if (fsc->new_version) server_write_char(fsc, "I");
	server_write_id(fsc, ioformat);
	break;
    case 'G':
	if (no_fork) printf("Get Format   -> ");
	ioformat = io_malloc(sizeof(IOFormatStruct));
	ioformat->field_list = NULL;
	ioformat->var_list = NULL;
	ioformat->field_count = 0;
	ioformat->format_name = NULL;
	ioformat->conversion = NULL;
	server_get_server_ID(fsc->fd, &ioformat->server_ID);
	if (no_fork) {
	    print_format_ID(ioformat);
	    printf("\n");
	}
	ioformat = find_format(fs, fsc, ioformat,
			     0 /* not new format mode */);
	if (no_fork) {
	    printf("Returning -> ");
	    if (ioformat == NULL) {
		printf("	NULL\n");
	    } else {
		dump_IOFormat(ioformat);
	    }
	}
	if (fsc->new_version) server_write_char(fsc, "F");
	write_format_to_fd(fsc->fd, ioformat);
	break;
    default:
	printf("Unknown request %c, %x\n", next_action, (unsigned char) next_action);
	break;
    }
    
}

/* 
 * Create and initialize a connection data structure.
 */
static FSClient
FSClient_create(format_server fs)
{
    FSClient fsc = (FSClient) io_malloc(sizeof(*fsc));
    fsc->fd = NULL;
    fsc->hostname = NULL;
    fsc->exch_pid = 0;
    fsc->port = 0;
    fsc->fs = fs;
    fsc->new_version = 0;
    return fsc;
}

static void
get_qual_hostname(char *buf, int len)
{
    gethostname(buf, len);
    buf[len-1] = '\0';
    if (memchr(buf, '.', strlen(buf)) == NULL) {
	 /* no dots, probably not fully qualified */
#ifdef HAVE_GETDOMAINNAME
	 int end = strlen(buf);
	 buf[end] = '.';
	 getdomainname((&buf[end])+1, len-strlen(buf));
#else
	 /* no getdomainname, hope that gethostbyname will help */
	 char *tmp_name = (gethostbyname(buf))->h_name;
	 strncpy(buf, tmp_name, len);
#endif
	 buf[len-1] = '\0';
    }
}

/* 
 * Create and initialize a data exchange.
 */
static format_server
format_server_create()
{
    int len, i, j;
    char name_buf[256];
    format_server fs = (format_server)io_malloc(sizeof(*fs));
    int max_fd = FD_SETSIZE;	/* returns process fd table size */
    if (fs == NULL)
	return NULL;

    /* initialize data structs */
    get_qual_hostname(name_buf, sizeof(name_buf));
    len = strlen(name_buf);
    name_buf[len] = '_';
    sprintf(&name_buf[len + 1], "%u", (unsigned int) getpid());
    fs->exchange_name = strdup(name_buf);
    fs->exch_pid = getpid();
    fs->port = -1;
    fs->hostname = NULL;
    fs->conn_sock_inet = (void *)-1;
    fs->usock_name = NULL;
    for(i = 0; i<2 ; i++) {
	for (j=0; j<1; j++) {
	    fs->lists[i][j] = NULL;
	}
    }

    fs->portCount = 0;
    fs->ports = (FSClient *) io_malloc(sizeof(FSClient) * max_fd);
    memset((char *) fs->ports, 0, sizeof(FSClient) * max_fd);

    FD_ZERO(&fs->fdset);
    fs->data_buffer = (char*) io_malloc(1);
    fs->buffer_size = 1;
    /* 
     * ignore SIGPIPE's  (these pop up when ports die.  we catch the 
     * failed writes) 
     */
#ifdef SIGPIPE
    signal(SIGPIPE, SIG_IGN);
#endif

    return fs;
}


/* 
 * Create an IP socket for connection from other data exchanges.
 */
static int
format_server_listen(format_server fs, int port_num)
{
    int length;
    char buf[256];
    struct sockaddr_in sock_addr;
    int sock_opt_val = 1;
    int conn_sock;

    if ((long)fs->conn_sock_inet >= 0) {
	fprintf(stderr, "INET connection socket already open %lx\n",
		(long)fs->conn_sock_inet);
	return -1;
    }
    conn_sock = socket(AF_INET, SOCK_STREAM, 0);
    if (conn_sock < 0) {
	
	fprintf(stderr, "Cannot open INET socket %d\n", conn_sock);
	return -1;
    }
    sock_addr.sin_family = AF_INET;
    sock_addr.sin_addr.s_addr = INADDR_ANY;
    sock_addr.sin_port = htons((unsigned short) port_num);
    if (setsockopt((long)conn_sock, SOL_SOCKET, SO_REUSEADDR, (char *) &sock_opt_val,
		   sizeof(sock_opt_val)) != 0) {
	fprintf(stderr, "Failed to set 1REUSEADDR on INET socket\n");
	return -1;
    }
    if (bind(conn_sock, (struct sockaddr *) &sock_addr,
	     sizeof sock_addr) < 0) {
	fprintf(stderr, "Cannot bind INET socket\n");
	return -1;
    }
    sock_opt_val = 1;
    if (setsockopt(conn_sock, SOL_SOCKET, SO_REUSEADDR, (char *) &sock_opt_val,
		   sizeof(sock_opt_val)) != 0) {
	perror("Failed to set 2REUSEADDR on INET socket");
	return -1;
    }
    length = sizeof sock_addr;
    if (getsockname(conn_sock, (struct sockaddr *) &sock_addr, &length) < 0) {
	fprintf(stderr, "Cannot get socket name\n");
	return -1;
    }
    fs->port = ntohs(sock_addr.sin_port);

    /* begin listening for conns and set the backlog */
    if (listen(conn_sock, FD_SETSIZE)) {
	fprintf(stderr, "listen failed\n");
	return -1;
    }
    /* record the server's hostname */
    get_qual_hostname(buf, sizeof(buf));
    fs->hostname = (char *) strdup(buf);

    fs->conn_sock_inet = (void *)conn_sock;
    if (conn_sock > FD_SETSIZE) {
	fprintf(stderr, "Internal Error, stupid WINSOCK large FD bug.\n");
	fprintf(stderr, "Increase FD_SETSIZE.  Item not added to fdset.\n");
    }
    FD_SET(conn_sock, &fs->fdset);
    return 0;
}

static void out_domain_rejection(int fd, FSClient fsc);

/* 
 * Accept socket connection from other data exchs.  If conn_sock is -1,
 * check if there is connection request at conn_sock_inet or
 * conn_sock_unix.
 */
extern FSClient
format_server_accept_conn_sock(format_server fs, void *conn_sock)
{
    int sock;
    FSClient fsc;

    struct linger linger_val;

    linger_val.l_onoff = 1;
    linger_val.l_linger = 60;
    if (fs == NULL)
	return NULL;
    if ((long)conn_sock == -1) {
	fd_set fds;
	struct timeval timeout;
	FD_ZERO(&fds);
	if ((long)fs->conn_sock_inet >= 0) {
	    FD_SET((unsigned long)fs->conn_sock_inet, &fds);
	    if ((long)fs->conn_sock_inet > FD_SETSIZE) {
		fprintf(stderr, "Internal Error, stupid WINSOCK large FD bug.\n");
		fprintf(stderr, "Increase FD_SETSIZE.  Item not added to fdset.\n");
	    }
	}
	timeout.tv_sec = 0L;
	timeout.tv_usec = 0L;

	if (select(FD_SETSIZE, &fds, (fd_set *) NULL, (fd_set *) NULL,
		   &timeout) > 0) {
	    if (FD_ISSET((int)(long)fs->conn_sock_inet, &fds)) {
		format_server_accept_conn_sock(fs, fs->conn_sock_inet);
	    }
	}
	return NULL;
    }
    if ((sock = accept((unsigned int)(long)conn_sock, (struct sockaddr *) 0, (int *) 0)) < 0) {
	fprintf(stderr, "Cannot accept socket connection\n");
	return NULL;
    }
    if (setsockopt(sock, SOL_SOCKET, SO_LINGER, (char *) &linger_val,
		   sizeof(struct linger)) != 0) {
	perror("set SO_LINGER");
	return NULL;
    }

    fsc = FSClient_create(fs);
    fsc->fd = (void*)sock;

    FD_SET(sock, &fs->fdset);
    if (sock > FD_SETSIZE) {
	fprintf(stderr, "Internal Error, stupid WINSOCK large FD bug.\n");
	fprintf(stderr, "Increase FD_SETSIZE.  Item not added to fdset.\n");
    }
    fs->ports[sock] = fsc;
    fs->portCount++;

    if (no_fork) {
	printf("%s: FSxchange_accept_conn_sock successful\n",
	       fs->exchange_name);
    }
    if (fsc->hostname == NULL) fsc->hostname = "none";
    if (no_fork) {
	printf("fsc->fd = %d, exch_pid %lx, port = %d, hostname = %s\n", 
	       (int)(long)fsc->fd, (long)fsc->exch_pid, fsc->port, 
	       fsc->hostname);
    }
    server_read_header(fsc);
    out_domain_rejection(sock, fsc);
    return fsc;
}

typedef struct host_info {
    struct in_addr ip_addr;
    time_t intro_time;
} *host_info_p;

static host_info_p hostlist = NULL;
static int host_count = 0;

static time_t
get_time_for_host(ip_addr, hostname)
struct in_addr ip_addr;
char *hostname;
{
    int i;
    if (hostlist == NULL) {
	hostlist = malloc(sizeof(struct host_info));
    }
    for (i = 0; i< host_count; i++) {
	if (memcmp(&ip_addr, &hostlist[i].ip_addr, sizeof(struct in_addr)) == 0) {
	    return hostlist[i].intro_time;
	}
    }
    hostlist = realloc(hostlist, sizeof(struct host_info) * (host_count + 1));
    memcpy(&hostlist[host_count].ip_addr, &ip_addr, sizeof(struct in_addr));
    hostlist[host_count].intro_time = time(NULL);
    host_count++;
    return hostlist[host_count-1].intro_time;
}

static char *postfix = FORMAT_SERVICE_DOMAIN;

/* 3 days of free use of our daemons for every host */
#define GRACE_PERIOD_SEC 60 * 60 * 24 * 3

static void
out_domain_rejection(fd, fsc)
int fd;
FSClient fsc;
{
    struct sockaddr sock_addr;
    int sock_len = sizeof(sock_addr);
    time_t intro_time;

    if (getpeername(fd, &sock_addr, &sock_len) == 0) {
	struct hostent *host;
	if (sock_addr.sa_family == AF_INET) {
	    struct sockaddr_in *in_sock = (struct sockaddr_in *) &sock_addr;
	    host = gethostbyaddr((char *) &in_sock->sin_addr,
				 sizeof(struct in_addr), AF_INET);
	    if (host != NULL) {
		char **alias;
		int postlen = strlen(postfix);
		int hostlen = strlen(host->h_name);
		/* 
		 * Check to see if our FORMAT_SERVICE_DOMAIN is a postfix
		 * of the hostname.  If not, we'll reject the host as being
		 * outside of our service domain.
		 */
		if (hostlen > postlen) {
		    if (strcmp(&host->h_name[hostlen - postlen], postfix) == 0) {
			/* good host */
			return;
		    }
		} else {
		    if (strcmp(host->h_name, "localhost") == 0) {
			return;
		    }
		}
		for (alias = host->h_aliases; *alias != 0; alias++) {
		    if (hostlen > postlen) {
			if (strcmp(alias[hostlen - postlen], postfix) == 0) {
			    /* good host */
			    return;
			}
		    }
		}

		intro_time = get_time_for_host(in_sock->sin_addr, host->h_name);
		if ((time(NULL) - intro_time) < GRACE_PERIOD_SEC) {
		    if (fsc->new_version) {
			printf("Sending character \"P\"\n");
			server_write_char(fsc, "P");
		    } else {
			close(fd);
		    }
		    return;
		}

		printf("Rejecting out-domain host %s, aliases: ", host->h_name);
		for (alias = host->h_aliases; *alias != 0; alias++) {
		    printf(" %s", *alias);
		}
		printf("\n");
	    }
	}
    }
    /* must not be a good host, close it down ... */
    close(fd);
}

