#include "config.h"
#include "assert.h"
#include <stdio.h>
#ifdef STDC_HEADERS
#include <stdlib.h>
#endif
#ifdef HAVE_MEMORY_H
#include <memory.h>
#endif
#ifdef HAVE_STRING_H
#include <string.h>
#endif
#include <sys/types.h>
#include <ctype.h>
#ifdef HAVE_SYS_UIO_H
#include <sys/uio.h>
#endif
#include "io.h"
#include "io_interface.h"
#include "io_internal.h"
#include "unix_defs.h"
#include "assert.h"
extern char *getenv ARGS((const char *name));

#ifdef HAVE_VCODE_H
#define __NDEBUG__
#include "vcode.h"
#include "io_gen.h"
#endif

static MAX_INTEGER_TYPE get_big_int ARGS((IOFieldPtr iofield, void *data));
static MAX_FLOAT_TYPE get_big_float ARGS((IOFieldPtr iofield, void *data));
static MAX_UNSIGNED_TYPE get_big_unsigned ARGS((IOFieldPtr iofield, void *data));
static void byte_swap ARGS((char *data, int size));

static int get_double_warn = 0;
static int get_long_warn = 0;

static void
convert_addr_field(iofile, src_spec, src, dest_size, dest, string_offset_size,
		   string_base_p, size_delta, converted_strings, src_offset_p,
		   dest_p, required_alignment)
IOFile iofile;
IOFieldPtr src_spec;
void *src;
int dest_size;
void *dest;
int string_offset_size;
char **string_base_p;
int size_delta;
int converted_strings;
int *src_offset_p;
void **dest_p;
int required_alignment;
{
    int align_tmp;
    if ((dest_size == sizeof(char *)) && (*string_base_p != NULL)) {
	IOgetFieldStruct tmp_src_field;
	MAX_INTEGER_TYPE tmp_int;
	char **dest_field = (char **) dest;
	tmp_src_field = *src_spec;

	tmp_src_field.data_type = integer_type;

	tmp_int = get_big_int(&tmp_src_field, src);

	*src_offset_p = tmp_int;
	if (tmp_int != 0) {
	    /* handle possibly different string base */
	    tmp_int -= (long) string_offset_size;
	    *dest_field = *string_base_p + tmp_int;
	    if ((align_tmp = (((unsigned long)*dest_field) % required_alignment)) != 0) {
		*dest_field += (required_alignment - align_tmp);
		*string_base_p  += (required_alignment - align_tmp);
	    }
	} else {
	    *dest_field = NULL;
	}
	*dest_p = *dest_field;
    } else if ((dest_size > sizeof(char *)) && (*string_base_p != NULL)) {
	/* native field is bigger than char*, store it */
	IOgetFieldStruct tmp_src_field;
	MAX_UNSIGNED_TYPE tmp_int;
	tmp_src_field = *src_spec;

	tmp_src_field.data_type = integer_type;

	tmp_int = get_big_unsigned(&tmp_src_field, src);

	*src_offset_p = tmp_int;

	if (tmp_int != 0) {
	    /* handle possibly different string base */
	    tmp_int -= (long) string_offset_size;
	    *dest_p = tmp_int + *string_base_p;
	    tmp_int = tmp_int + (MAX_UNSIGNED_TYPE) (unsigned long) *string_base_p;
	    if ((align_tmp = (((unsigned long)*dest_p) % required_alignment)) != 0) {
		*dest_p = (char*)*dest_p + (required_alignment - align_tmp);
		tmp_int += (required_alignment - align_tmp);
		*string_base_p  += (required_alignment - align_tmp);
	    }
	} else {
	    *dest_p = NULL;
	}
	tmp_src_field.offset = 0;
	tmp_src_field.size = sizeof(MAX_UNSIGNED_TYPE);
	tmp_src_field.byte_swap = FALSE;
	pbio_internal_convert_field(iofile, &tmp_src_field, &tmp_int,
				    unsigned_type, dest_size,
				    dest, 0, *string_base_p, 0, FALSE);
    } else {
	/* not a native field struct.  Keep the pointer as an offset. */
	IOgetFieldStruct tmp_src_field;
	MAX_INTEGER_TYPE tmp_int;
	tmp_src_field = *src_spec;

	tmp_src_field.data_type = integer_type;
	tmp_int = get_big_int(&tmp_src_field, src);

	*src_offset_p = tmp_int;
	if (tmp_int != 0) {
	    if (!converted_strings) {
		/* record size might change, adjust offset */
		tmp_int += size_delta;
	    }
	    *dest_p = tmp_int + *string_base_p;
	    if ((align_tmp = (((unsigned long)*dest_p) % required_alignment)) != 0) {
		*dest_p =  (char*)*dest_p + (required_alignment - align_tmp);
		tmp_int += (required_alignment - align_tmp);
		*string_base_p  += (required_alignment - align_tmp);
	    }
	} else {
	    *dest_p = NULL;
	}
	tmp_src_field.offset = 0;
	tmp_src_field.size = sizeof(MAX_INTEGER_TYPE);
	tmp_src_field.byte_swap = FALSE;
	pbio_internal_convert_field(iofile, &tmp_src_field, &tmp_int,
				    integer_type, dest_size,
				    dest, 0, *string_base_p, 0, FALSE);
    }
}

extern void
pbio_internal_convert_field(iofile, src_spec, src, dest_type, dest_size, dest,
	  string_offset_size, string_base, size_delta, converted_strings)
IOFile iofile;
IOFieldPtr src_spec;
void *src;
IOdata_type dest_type;
int dest_size;
void *dest;
int string_offset_size;
char *string_base;
int size_delta;
int converted_strings;
{
    /* quick check to see if it's just a copy... */
    if ((dest_type != string_type) && (dest_type == src_spec->data_type) &&
	(dest_size == src_spec->size)) {
	if (src_spec->byte_swap) {
	    int i;
	    char *destc = (char *) dest;
	    char *srcc = (char *) src + src_spec->offset;
	    for (i = 0; i < dest_size / 2; i++) {
		char tmp = srcc[dest_size - i - 1];
		destc[dest_size - i - 1] = srcc[i];
		destc[i] = tmp;
	    }
	    if (dest_size % 2 != 0) {
		destc[dest_size / 2] = srcc[dest_size / 2];
	    }
	} else {
	    char *srcc = (char *) src + src_spec->offset;
	    memcpy(dest, srcc, dest_size);
	}
	return;
    }
    switch (dest_type) {
    case integer_type:
	{
	    MAX_INTEGER_TYPE tmp = get_big_int(src_spec, src);
	    if (dest_size == sizeof(char)) {
		char *dest_field = (char *) dest;
		*dest_field = (char) tmp;
	    } else if (dest_size == sizeof(short)) {
		short *dest_field = (short *) dest;
		*dest_field = (short) tmp;
	    } else if (dest_size == sizeof(int)) {
		int *dest_field = (int *) dest;
		*dest_field = tmp;
	    } else if (dest_size == sizeof(long)) {
		long *dest_field = (long *) dest;
		*dest_field = tmp;
#if SIZEOF_LONG_LONG != 0
	    } else if (dest_size == sizeof(long long)) {
		long long lltmp = tmp;
		memcpy(dest, &lltmp, sizeof(long long));
#endif
	    } else {
		iofile->result = "size problems in conversion";
	    }
	    break;
	}
    case char_type:
	{
	    char tmp = get_IOchar(src_spec, src);
	    char *dest_field = (char *) dest;
	    *dest_field = tmp;
	    break;
	}
    case enumeration_type:
	{
	    int tmp = get_IOenum(src_spec, src);
	    int *dest_field = (int *) dest;
	    *dest_field = tmp;
	    break;
	}
    case unsigned_type:
	{
	    MAX_UNSIGNED_TYPE tmp = get_big_unsigned(src_spec, src);
	    if (dest_size == sizeof(unsigned char)) {
		unsigned char *dest_field = (unsigned char *) dest;
		*dest_field = (unsigned char) tmp;
	    } else if (dest_size == sizeof(unsigned short)) {
		unsigned short *dest_field = (unsigned short *) dest;
		*dest_field = (unsigned short) tmp;
	    } else if (dest_size == sizeof(int)) {
		unsigned int *dest_field = (unsigned int *) dest;
		*dest_field = tmp;
	    } else if (dest_size == sizeof(long)) {
		unsigned long *dest_field = (unsigned long *) dest;
		*dest_field = tmp;
#if SIZEOF_LONG_LONG != 0
	    } else if (dest_size == sizeof(long long)) {
		unsigned long long *dest_field = (unsigned long long *) dest;
		*dest_field = tmp;
#endif
	    } else {
		iofile->result = "size problems in conversion";
	    }
	    break;
	}
    case float_type:
	{
	    MAX_FLOAT_TYPE tmp = get_big_float(src_spec, src);
	    if (dest_size == sizeof(float)) {
		float *dest_field = (float *) dest;
		*dest_field = (float) tmp;
	    } else if (dest_size == sizeof(double)) {
		double *dest_field = (double *) dest;
		*dest_field = tmp;
#if SIZEOF_LONG_DOUBLE != 0
	    } else if (dest_size == sizeof(long double)) {
		long double *dest_field = (long double *) dest;
		*dest_field = tmp;
#endif
	    } else {
		iofile->result = "size problems in conversion";
	    }
	    break;
	}
    case string_type:
	{
	    void *junk;
	    int junk_int;
	    /* use Junk values because we don't care about address output */
	    convert_addr_field(iofile, src_spec, src, dest_size, dest,
			       string_offset_size, &string_base, size_delta,
			       converted_strings, &junk_int, &junk, 1);
	    break;
	}
    default:
	assert(FALSE);
    }
}

static void
do_var_part_conv ARGS((IOConversionPtr conv, IOconvFieldStruct * conv_field,
		     void *src_area, void *final_area, int control_value,
		       void *src_string_base, void **final_string_base));

static int debug_code_generation = -1;

void
IOconvert_record(conv, src, dest, final_string_base, src_string_base)
IOConversionPtr conv;
void *src;
void *dest;
void *final_string_base;
void *src_string_base;
{
    int i;
    long *control_value = NULL;

    if (src_string_base == NULL) {
	src_string_base = final_string_base;
    }
    if (conv->conv_func) {
	if (debug_code_generation) {
	    int i;
	    int *tmp = (int *) (((char *) src_string_base) -
				(((long) src_string_base) % 4));
	    printf("record contents :\n");
	    for (i = 0; i < 30; i += 4) {
		printf("%lx: %8x %8x %8x %8x\n", (long) ((char *) src) + (i * 4),
		       ((int *) src)[i], ((int *) src)[i + 1],
		       ((int *) src)[i + 2], ((int *) src)[i + 3]);
	    }
	    if (src_string_base != NULL) {
		printf("string contents :\n");
		for (i = 0; i < 30; i += 4) {
		    printf("%lx: %8x %8x %8x %8x\n", (long) ((char *) tmp) + (i * 4),
			   ((int *) tmp)[i],
			   ((int *) tmp)[i + 1],
			   ((int *) tmp)[i + 2],
			   ((int *) tmp)[i + 3]);
		}
	    }
	}
	conv->conv_func(src, dest, final_string_base, src_string_base);
	return;
    }
    for (i = 0; i < conv->conv_count; i++) {
	if (conv->conversions[i].control_field != NULL) {
	    if (control_value == NULL) {
		int j;
		control_value = (long *) io_malloc(sizeof(long) * conv->conv_count);
		for (j = 0; j < conv->conv_count; j++)
		    control_value[j] = 0;
	    }
	    control_value[i] = (long)
		get_big_int(conv->conversions[i].control_field, src);
	}
    }
    for (i = 0; i < conv->conv_count; i++) {
	IOFieldPtr src_spec = &conv->conversions[i].src_field;
	int dimen1 = conv->conversions[i].array_dimen1;
	int dimen2 = conv->conversions[i].array_dimen2;

	if (IOhas_error(conv->iofile))
	    return;
	if (!conv->conversions[i].src_field.byte_swap &&
	    (src_spec->size == conv->conversions[i].dest_size) &&
	    (conv->conversions[i].subconversion == NULL) &&
	    (conv->conversions[i].control_field == NULL) &&
	    ((src_spec->data_type != string_type) || (final_string_base == NULL))) {
	    /* data movement is all that is required */
	    void *dest_field = (void *) (conv->conversions[i].dest_offset +
					 (char *) dest);
	    memcpy(dest_field, (char *) src + src_spec->offset,
		   src_spec->size * dimen1 * dimen2);
	} else {
	    int j;
	    void *dest_field = (void *) (conv->conversions[i].dest_offset +
					 (char *) dest);
	    IOgetFieldStruct tmp_spec;
	    IOdata_type dest_type = src_spec->data_type;
	    int dest_size = conv->conversions[i].dest_size;
	    tmp_spec = *src_spec;

	    if (conv->conversions[i].control_field != NULL) {
		/* really a variant array, adjust address like a string */
		tmp_spec.data_type = string_type;
		tmp_spec.size = conv->ioformat->pointer_size;
		dest_type = string_type;
		dest_size = conv->target_pointer_size;
	    }
	    for (j = 0; j < dimen1 * dimen2; j++) {
		tmp_spec.offset = src_spec->offset + j * tmp_spec.size;
		if (dest_type != unknown_type) {
		    /* simple (native) field or variant array */
		    if (dest_type != string_type) {
			pbio_internal_convert_field(conv->iofile, &tmp_spec, src,
						    dest_type, dest_size,
				    dest_field, conv->string_offset_size,
						    final_string_base,
						    conv->base_size_delta,
						conv->converted_strings);
		    } else {
			int src_offset;
			void *src_base;
			void *dst_base;
			/* string or variant array */
			int req_align;
			if (conv->conversions[i].control_field != NULL) {
			    req_align = min_align_size(conv->conversions[i].dest_size);
			} else {
			    req_align = 1;
			}
			convert_addr_field(conv->iofile, &tmp_spec, src,
					   dest_size, dest_field,
					   conv->string_offset_size,
					   &final_string_base,
					   conv->base_size_delta,
					   conv->converted_strings,
					   &src_offset, &dst_base,
					   req_align);
			src_base = (char *) src_string_base + src_offset
			    - conv->string_offset_size;
			if ((conv->conversion_type == copy_strings) ||
			  (conv->conversions[i].control_field != NULL)) {
			    int control = 0;
			    if (control_value != NULL)
				control = control_value[i];
			    do_var_part_conv(conv, &conv->conversions[i],
					     src_base, dst_base,
					     control, src_string_base,
					     &final_string_base);
			}
		    }
		} else {
		    IOConversionPtr subconv =
		    conv->conversions[i].subconversion;
		    void *subsrc = (void *) ((char *) src + tmp_spec.offset);
		    IOconvert_record(subconv, subsrc, dest_field,
				     final_string_base, src_string_base);
		}
		dest_field = (void *) ((char *) dest_field +
				       conv->conversions[i].dest_size);
	    }
	}
    }
    if (control_value != NULL) {
	io_free(control_value);
    }
}

static void
do_var_part_conv(conv, conv_field, src_area, final_area,
		 control_value, src_string_base, final_string_base)
IOConversionPtr conv;
IOconvFieldStruct *conv_field;
void *src_area;
void *final_area;
int control_value;
void *src_string_base;
void **final_string_base;
{
    int copy_length = 0;
    IOFieldPtr src_spec = &(conv_field->src_field);

    /* handle copying or conversion */
    if (conv_field->src_field.data_type == string_type) {
	if (src_area != NULL) {
	    int copy_length = strlen(src_area) + 1;
	    memcpy(final_area, src_area, copy_length);
	}
    } else {
	/* variant array */
	if (!conv_field->src_field.byte_swap &&
	    (src_spec->size == conv_field->dest_size) &&
	    (conv_field->subconversion == NULL) &&
	    (src_spec->data_type != string_type)) {

	    /* data movement is all that is required */
	    if (conv->conversion_type == copy_strings) {
		copy_length = conv_field->dest_size *
		    control_value;
		memcpy(final_area, src_area, copy_length);
	    } else {
		/* not even that is necessary */
	    }
	} else {
	    /* must do conversions */
	    int length = control_value;
	    int j;
	    void *dest_field = final_area;
	    IOgetFieldStruct tmp_spec;
	    IOdata_type dest_type = src_spec->data_type;
	    int dest_size = conv_field->dest_size;

	    tmp_spec = *src_spec;
	    *final_string_base = (char *) *final_string_base +
		(length * (dest_size - src_spec->size));
	    for (j = 0; j < length; j++) {
		tmp_spec.offset = j * src_spec->size;
		if (dest_type != unknown_type) {
		    /* simple (native) field in variant array */
		    pbio_internal_convert_field(conv->iofile, &tmp_spec,
					  src_area, dest_type, dest_size,
						dest_field,
						conv->string_offset_size,
						*final_string_base,
						conv->base_size_delta,
						conv->converted_strings);
		} else {
		    IOConversionPtr subconv =
		    conv_field->subconversion;
		    void *subsrc = (char *) src_area + j * src_spec->size;
		    IOconvert_record(subconv, subsrc, dest_field,
				     *final_string_base,
				     src_string_base);
		}
		dest_field = (void *) ((char *) dest_field +
				       conv_field->dest_size);
	    }
	}
    }
}

static MAX_INTEGER_TYPE
get_big_int(iofield, data)
IOFieldPtr iofield;
void *data;
{
    if (iofield->data_type == integer_type) {
	if (iofield->size == sizeof(char)) {
	    char tmp;
	    memcpy(&tmp, (char *) data + iofield->offset, sizeof(char));
	    return (long) tmp;
	} else if (iofield->size == sizeof(short)) {
	    short tmp;
	    memcpy(&tmp, (char *) data + iofield->offset, sizeof(short));
	    if (iofield->byte_swap)
		byte_swap((char *) &tmp, sizeof(short));
	    return (long) tmp;
	} else if (iofield->size == sizeof(int)) {
	    int tmp;
	    memcpy(&tmp, (char *) data + iofield->offset, sizeof(int));
	    if (iofield->byte_swap)
		byte_swap((char *) &tmp, sizeof(int));
	    return (long) tmp;
	} else if (iofield->size == sizeof(long)) {
	    long tmp;
	    memcpy(&tmp, (char *) data + iofield->offset, sizeof(long));
	    if (iofield->byte_swap)
		byte_swap((char *) &tmp, sizeof(long));
	    return tmp;
	} else if (iofield->size == 2 * sizeof(long)) {
	    long tmp;
	    int low_bytes_offset = iofield->offset;
#ifdef WORDS_BIGENDIAN
	    if (!iofield->byte_swap) {
		low_bytes_offset += sizeof(long);
	    }
#else
	    if (iofield->byte_swap) {
		low_bytes_offset += sizeof(long);
	    }
#endif
	    memcpy(&tmp, (char *) data + low_bytes_offset, sizeof(long));
	    if (iofield->byte_swap)
		byte_swap((char *) &tmp, sizeof(long));
	    return tmp;
	} else {
	    if (!IO_shut_up && !get_long_warn) {
		fprintf(stderr, "Get Long failed!  Size problems.  File int size is %d.\n", iofield->size);
		get_long_warn++;
	    }
	    return -1;
	}
    } else if (iofield->data_type == unsigned_type) {
	MAX_UNSIGNED_TYPE tmp = get_big_unsigned(iofield, data);
	return (MAX_UNSIGNED_TYPE) tmp;
    } else if (iofield->data_type == float_type) {
	MAX_FLOAT_TYPE tmp = get_big_float(iofield, data);
#ifndef METICULOUS_FLOATS_AND_LONGS
	return (MAX_INTEGER_TYPE) (long) (double) tmp;
#else
	return (MAX_INTEGER_TYPE) tmp;
#endif
    } else {
	fprintf(stderr, "Get IOlong failed on invalid data type!\n");
	exit(1);
    }
    /* NOTREACHED */
    return 0;
}

static MAX_UNSIGNED_TYPE
get_big_unsigned(iofield, data)
IOFieldPtr iofield;
void *data;
{
    if (iofield->data_type == unsigned_type) {
	if (iofield->size == sizeof(char)) {
	    unsigned char tmp;
	    memcpy(&tmp, (char *) data + iofield->offset, sizeof(char));
	    return (MAX_UNSIGNED_TYPE) tmp;
	} else if (iofield->size == sizeof(short)) {
	    unsigned short tmp;
	    memcpy(&tmp, (char *) data + iofield->offset, sizeof(short));
	    if (iofield->byte_swap)
		byte_swap((char *) &tmp, sizeof(short));
	    return (MAX_UNSIGNED_TYPE) tmp;
	} else if (iofield->size == sizeof(int)) {
	    unsigned int tmp;
	    memcpy(&tmp, (char *) data + iofield->offset, sizeof(int));
	    if (iofield->byte_swap)
		byte_swap((char *) &tmp, sizeof(int));
	    return (MAX_UNSIGNED_TYPE) tmp;
	} else if (iofield->size == sizeof(long)) {
	    unsigned long tmp;
	    memcpy(&tmp, (char *) data + iofield->offset, sizeof(long));
	    if (iofield->byte_swap)
		byte_swap((char *) &tmp, sizeof(long));
	    return tmp;
	} else if (iofield->size == 2 * sizeof(long)) {
	    unsigned long tmp;
	    int low_bytes_offset = iofield->offset;
#ifdef WORDS_BIGENDIAN
	    if (!iofield->byte_swap) {
		low_bytes_offset += sizeof(long);
	    }
#else
	    if (iofield->byte_swap) {
		low_bytes_offset += sizeof(long);
	    }
#endif
	    memcpy(&tmp, (char *) data + low_bytes_offset, sizeof(long));
	    if (iofield->byte_swap)
		byte_swap((char *) &tmp, sizeof(long));
	    return tmp;
	} else {
	    if (!IO_shut_up && !get_long_warn) {
		fprintf(stderr, "Get Long failed!  Size problems.  File int size is %d.\n", iofield->size);
		get_long_warn++;
	    }
	    return 0;
	}
    } else if (iofield->data_type == integer_type) {
	MAX_INTEGER_TYPE tmp = get_big_int(iofield, data);
	return (MAX_UNSIGNED_TYPE) tmp;
    } else if (iofield->data_type == float_type) {
	MAX_FLOAT_TYPE tmp = get_big_float(iofield, data);
#ifndef METICULOUS_FLOATS_AND_LONGS
	return (MAX_UNSIGNED_TYPE) (long) (double) tmp;
#else
	return (MAX_UNSIGNED_TYPE) tmp;
#endif
    } else {
	fprintf(stderr, "Get IOulong failed on invalid data type!\n");
	exit(1);
    }
    /* NOTREACHED */
    return 0;
}

static MAX_FLOAT_TYPE
get_big_float(iofield, data)
IOFieldPtr iofield;
void *data;
{
    if (iofield->data_type == float_type) {
	if (iofield->size == sizeof(float)) {
	    float tmp;
	    MAX_FLOAT_TYPE tmp2;
	    memcpy(&tmp, ((char *) data + iofield->offset), sizeof(float));
	    if (iofield->byte_swap)
		byte_swap((char *) &tmp, sizeof(float));
	    tmp2 = tmp;
	    return tmp2;
	} else if (iofield->size == sizeof(double)) {
	    double tmp;
	    MAX_FLOAT_TYPE tmp2;
	    memcpy(&tmp, ((char *) data + iofield->offset), sizeof(double));
	    if (iofield->byte_swap)
		byte_swap((char *) &tmp, sizeof(double));
	    tmp2 = tmp;
	    return tmp2;
#if SIZEOF_LONG_DOUBLE != 0
	} else if (iofield->size == sizeof(long double)) {
	    long double tmp;
	    MAX_FLOAT_TYPE tmp2;
	    memcpy(&tmp, ((char *) data + iofield->offset),
		   sizeof(long double));
	    if (iofield->byte_swap)
		byte_swap((char *) &tmp,
			  sizeof(long double));
	    tmp2 = tmp;
	    return tmp2;
#endif
	} else {
	    if (!IO_shut_up && !get_double_warn) {
		fprintf(stderr, "Get Double failed!  Size problems.  File double size is %d.\n", iofield->size);
		get_double_warn++;
	    }
	    return 0.0;
	}
    } else if (iofield->data_type == integer_type) {
	MAX_INTEGER_TYPE tmp = get_big_int(iofield, data);
#ifndef METICULOUS_FLOATS_AND_LONGS
	/* 
	 * A concession to inter-compiler interoperability...  We
	 * shouldn't need the (double)(long) casts here.  If we don't
	 * use them, AND we're on a machine which doesn't support "long 
	 * long" and "long double" in native code, gcc generates calls 
	 * to libgcc and the resulting library can't be linked with
	 * anything but GCC.  This can be a problem for installed
	 * libraries.  Using the casts avoids that particular problem,
	 * at a cost of data loss in the case of someone converting a
	 * "long long" that doesn't fit in a long to a floating point
	 * value. 
	 */
	return (MAX_FLOAT_TYPE) (double) (long) tmp;
#else
	return (MAX_FLOAT_TYPE) tmp;
#endif
    } else if (iofield->data_type == unsigned_type) {
	MAX_UNSIGNED_TYPE tmp = get_big_unsigned(iofield, data);
#ifndef METICULOUS_FLOATS_AND_LONGS
	return (MAX_FLOAT_TYPE) (double) (long) tmp;
#else
	return (MAX_FLOAT_TYPE) tmp;
#endif
    } else {
	fprintf(stderr, "Get Double failed on invalid data type!\n");
	exit(1);
    }
    /* NOTREACHED */
    return 0;
}

static void
byte_swap(data, size)
char *data;
int size;
{
    int i;
    assert((size % 2) == 0);
    for (i = 0; i < size / 2; i++) {
	char tmp = data[i];
	data[i] = data[size - i - 1];
	data[size - i - 1] = tmp;
    }
}

#ifndef HAVE_VCODE_H
extern
 conv_routine
generate_conversion(conv)
IOConversionPtr conv;
{
    return NULL;
}
#else
static
 conv_routine
 generate_conversion_code ARGS((IOConversionPtr conv, v_reg_t * args,
				int assume_align));

static int pbio_conversion_generation = -1;
static int generation_verbose = -1;
void (*v_dump_routine)ARGS((v_code *routine));

extern
 conv_routine
generate_conversion(conv)
IOConversionPtr conv;
{
    void *insns;
    v_vptr conversion_routine;
    v_reg_t args[6];
    char *format_name = conv->ioformat->format_name;

    if (pbio_conversion_generation == -1) {
	pbio_conversion_generation =
	    (getenv("PBIO_CONVERSION_GENERATION") != NULL);
	debug_code_generation =
	    (getenv("PBIO_CONVERSION_DEBUG") != NULL);
	generation_verbose =
	    (getenv("PBIO_CONVERSION_VERBOSE") != NULL);
    }
    if (!pbio_conversion_generation)
	return NULL;
    if (generation_verbose) {
	printf("For format %s ===================\n", format_name);
	dump_IOConversion(conv);
    }
    insns = io_malloc(2048);
    v_lambda("convert", "%p%p%p%p", args, V_NLEAF, insns, 2048);
    if (debug_code_generation) {
	v_scallv((v_vptr) printf, "%P%P%p%p%p%p",
		 "convert for %s called with src= %lx, dest %lx, final_string =%lx, src_string =%lx\n",
		 format_name, args[0], args[1], args[2], args[3]);
    }
    generate_conversion_code(conv, args, TRUE /* assume_aligned */ );
    conversion_routine = v_end(0).v;
    if (generation_verbose) {
	if (v_dump_routine != NULL) {
	    v_dump_routine((v_code *) conversion_routine);
	} else {
	    printf("Call vdump_init() to enable dumping\n");
	}
    }
    return (conv_routine) conversion_routine;
}
/* #define REG_DEBUG(x) printf x ; */
#define REG_DEBUG(x)

static void
gen_var_part_conv ARGS((IOConversionPtr conv, int i, int control_base,
		   int assume_align, v_reg_t src_addr, v_reg_t dest_addr,
		    v_reg_t src_string_base, v_reg_t final_string_base));


static void
gen_simple_field_conv(tmp_spec, assume_align, src_addr, src_offset, dest_size,
		   dest_type, dest_addr, dest_offset, string_offset_size,
		   final_string_base, base_size_delta, converted_strings)
IOgetFieldStruct tmp_spec;
int assume_align;
v_reg_t src_addr;
int src_offset;
int dest_size;
IOdata_type dest_type;
v_reg_t dest_addr;
int dest_offset;
int string_offset_size;
v_reg_t final_string_base;
int base_size_delta;
int converted_strings;
{
    /* simple conversion */
    iogen_oprnd src_oprnd;
    int src_is_aligned = assume_align & ((src_offset % 8) == 0);
    int dst_is_aligned = assume_align & ((dest_offset % 8) == 0);

    assert(dest_type != string_type);

    src_oprnd = gen_operand(src_addr, src_offset, tmp_spec.size,
			    tmp_spec.data_type,
			    src_is_aligned,
			    tmp_spec.byte_swap);

    if (src_oprnd.address) {
	/* load the data if gen_byte_swap didn't */
	gen_load(&src_oprnd);
    }
    if (src_oprnd.data_type != dest_type) {
	iogen_oprnd tmp_oprnd;
	tmp_oprnd = gen_type_conversion(src_oprnd, dest_type);
	free_oprnd(src_oprnd);
	src_oprnd = tmp_oprnd;
    }
    if (src_oprnd.size != dest_size) {
	iogen_oprnd tmp_oprnd;
	tmp_oprnd = gen_size_conversion(src_oprnd, dest_size);
	free_oprnd(src_oprnd);
	src_oprnd = tmp_oprnd;
    }
    gen_store(src_oprnd, dest_addr, dest_offset,
	      dest_size, dest_type, dst_is_aligned);
    free_oprnd(src_oprnd);
}

static void
gen_addr_field_conv(tmp_spec, assume_align, src_addr, src_offset,
		    dest_size, dest_addr, dest_offset, string_offset_size,
  final_string_base, src_string_base, base_size_delta, converted_strings,
		    string_src_reg, string_dest_reg)
IOgetFieldStruct tmp_spec;
int assume_align;
v_reg_t src_addr;
int src_offset;
int dest_size;
v_reg_t dest_addr;
int dest_offset;
int string_offset_size;
v_reg_t final_string_base;
v_reg_t src_string_base;
int base_size_delta;
int converted_strings;
v_reg_t *string_src_reg;
v_reg_t *string_dest_reg;
{
    iogen_oprnd src_oprnd;
    int src_is_aligned = assume_align & ((src_offset % 8) == 0);
    int dst_is_aligned = assume_align & ((dest_offset % 8) == 0);

    v_label_t null_target = v_genlabel();
    if (!v_getreg(string_src_reg, V_P, V_TEMP))
	v_fatal("gen field convert out of registers \n");
    if (!v_getreg(string_dest_reg, V_P, V_TEMP))
	v_fatal("gen field convert out of registers \n");


    /* handle addresses */
    src_oprnd = gen_fetch(src_addr, src_offset, tmp_spec.size,
			  integer_type, src_is_aligned,
			  tmp_spec.byte_swap);

    /* src_oprnd now holds the offset value */
    if (dest_size >= sizeof(char *)) {

	if (src_oprnd.size != dest_size) {
	    /* make it the right size to operate on */
	    iogen_oprnd tmp_oprnd;
	    tmp_oprnd = gen_size_conversion(src_oprnd, sizeof(long));
	    free_oprnd(src_oprnd);
	    src_oprnd = tmp_oprnd;
	}
	/* generate : if it's zero, leave it zero  branch away */
	v_beqli(src_oprnd.vc_reg, 0, null_target);

	/* else, sub the string_offset_size */
	v_subli(src_oprnd.vc_reg, src_oprnd.vc_reg,
		string_offset_size);

	/* calculate the address of this in the source */
	v_addl(*string_src_reg, src_oprnd.vc_reg, src_string_base);

	/* and the address in the destination */
	v_addl(src_oprnd.vc_reg, src_oprnd.vc_reg,
	       final_string_base);

	v_label(null_target);

	v_movp(*string_dest_reg, src_oprnd.vc_reg);

	if (dest_size > sizeof(char *)) {
	    iogen_oprnd tmp_oprnd;
	    tmp_oprnd = gen_size_conversion(src_oprnd,
					    dest_size);
	    free_oprnd(src_oprnd);
	    src_oprnd = tmp_oprnd;
	}
	/* store the final address in the field */
	gen_store(src_oprnd, dest_addr, dest_offset, dest_size,
		  integer_type, dst_is_aligned);
	free_oprnd(src_oprnd);
	REG_DEBUG(("Regs %d and %d are src and dest \n",
		   _vrr(*string_src_reg), _vrr(*string_dest_reg)));
    } else {
	if (src_oprnd.size != dest_size) {
	    iogen_oprnd tmp_oprnd;
	    tmp_oprnd = gen_size_conversion(src_oprnd, dest_size);
	    free_oprnd(src_oprnd);
	    src_oprnd = tmp_oprnd;
	}
	/* dest field too small to store pointer in */
	/* store as integer offset */
	gen_store(src_oprnd, dest_addr, 0, dest_size,
		  integer_type, dst_is_aligned);

	/* generate : if it's zero, leave it zero  branch away */
	v_beqli(src_oprnd.vc_reg, 0, null_target);

	/* else, sub the string_offset_size */
	v_subli(src_oprnd.vc_reg, src_oprnd.vc_reg,
		string_offset_size);

	/* calculate the address of this in the source */
	v_addl(*string_src_reg, src_oprnd.vc_reg, src_string_base);

	/* and the address in the destination */
	v_addl(src_oprnd.vc_reg, src_oprnd.vc_reg,
	       final_string_base);

	v_label(null_target);

	v_movp(*string_dest_reg, src_oprnd.vc_reg);

	gen_store(src_oprnd, dest_addr, dest_offset, dest_size,
		  integer_type, dst_is_aligned);
	free_oprnd(src_oprnd);
    }
}

static void
gen_subfield_conv(conv, i, tmp_spec, dest_type, assume_align, src_addr,
	     src_offset, dest_size, dest_addr, dest_offset, control_base,
		  final_string_base, src_string_base)
IOConversionPtr conv;
int i;
IOgetFieldStruct tmp_spec;
IOdata_type dest_type;
int assume_align;
v_reg_t src_addr;
int src_offset;
int dest_size;
v_reg_t dest_addr;
int dest_offset;
int control_base;
v_reg_t final_string_base;
v_reg_t src_string_base;
{
    if (dest_type != unknown_type) {
	/* simple (native) field or variant array */
	if (dest_type != string_type) {
	    gen_simple_field_conv(tmp_spec, assume_align,
				  src_addr, src_offset,
				  dest_size, dest_type,
				  dest_addr, dest_offset,
				  conv->string_offset_size,
				  final_string_base,
				  conv->base_size_delta,
				  conv->converted_strings);
	} else {
	    v_reg_t string_src_reg;
	    v_reg_t string_dest_reg;
	    gen_addr_field_conv(tmp_spec, assume_align,
				src_addr, src_offset,
				dest_size, dest_addr,
				dest_offset,
				conv->string_offset_size,
				final_string_base,
				src_string_base,
				conv->base_size_delta,
				conv->converted_strings,
				&string_src_reg,
				&string_dest_reg);
	    if ((conv->conversion_type == copy_strings) ||
		(conv->conversions[i].control_field != NULL)) {
		gen_var_part_conv(conv, i, control_base,
				  assume_align,
				  string_dest_reg,
				  string_src_reg,
				  src_string_base,
				  final_string_base);
	    }
	    v_putreg(string_src_reg, V_P);
	    v_putreg(string_dest_reg, V_P);
	}
    } else {
	/* subconversion */
	IOConversionPtr subconv =
	conv->conversions[i].subconversion;
	v_reg_t new_src, new_dest;
	if (!v_getreg(&new_src, V_P, V_TEMP) ||
	    !v_getreg(&new_dest, V_P, V_TEMP))
	    v_fatal("temp vals in subcall\n");
	v_addpi(new_src, src_addr, src_offset);
	v_addpi(new_dest, dest_addr, dest_offset);
	v_scalli((v_iptr) subconv->conv_func, "%p%p%p%p", new_src,
		 new_dest, final_string_base, src_string_base);
	v_putreg(new_src, V_P);
	v_putreg(new_dest, V_P);
    }
}

extern
 conv_routine
generate_conversion_code(conv, args, assume_align)
IOConversionPtr conv;
v_reg_t *args;
int assume_align;
{
    int i;
    v_reg_t src_addr = args[0];
    v_reg_t dest_addr = args[1];
    v_reg_t final_string_base = args[2];
    v_reg_t src_string_base = args[3];
    int control_base = -1;

/* generate this    if (src_string_base == NULL) { src_string_base =
 * final_string_base; } */
    for (i = 0; i < conv->conv_count; i++) {
	iogen_oprnd src_oprnd;
	if (conv->conversions[i].control_field != NULL) {
	    IOFieldPtr control_field = conv->conversions[i].control_field;
	    int src_is_aligned = assume_align &
	    ((control_field->offset % 8) == 0);
	    if (control_base == -1) {
		control_base = v_localb(sizeof(int) * conv->conv_count);
	    }
	    src_oprnd = gen_fetch(src_addr, control_field->offset,
				  control_field->size,
				  control_field->data_type,
			       src_is_aligned, control_field->byte_swap);
	    if (src_oprnd.size != sizeof(int)) {
		iogen_oprnd tmp_oprnd;
		tmp_oprnd = gen_size_conversion(src_oprnd, sizeof(int));
		free_oprnd(src_oprnd);
		src_oprnd = tmp_oprnd;
	    }
	    gen_store(src_oprnd, v_lp, control_base + i * sizeof(int),
		      sizeof(int), integer_type, TRUE /* aligned */ );
	    free_oprnd(src_oprnd);
	}
    }
    for (i = 0; i < conv->conv_count; i++) {
	IOFieldPtr src_spec = &conv->conversions[i].src_field;
	int dimen1 = conv->conversions[i].array_dimen1;
	int dimen2 = conv->conversions[i].array_dimen2;
	if (!conv->conversions[i].src_field.byte_swap &&
	    (src_spec->size == conv->conversions[i].dest_size) &&
	    (conv->conversions[i].subconversion == NULL) &&
	    (conv->conversions[i].control_field == NULL) &&
	    ((src_spec->data_type != string_type))) {
	    /* data movement is all that is required */
	    int total_size = src_spec->size * dimen1 * dimen2;

	    if (total_size <= sizeof(long)) {
		iogen_oprnd src_oprnd;
		int src_is_aligned = assume_align &
		((src_spec->offset % 8) == 0);
		int dst_is_aligned = assume_align &
		((conv->conversions[i].dest_offset % 8) == 0);
		src_oprnd = gen_fetch(src_addr, src_spec->offset,
				      total_size, src_spec->data_type,
				      src_is_aligned, 0);
		gen_store(src_oprnd, dest_addr,
			  conv->conversions[i].dest_offset,
			  conv->conversions[i].dest_size,
			  src_spec->data_type, dst_is_aligned);
		free_oprnd(src_oprnd);
	    } else {
		v_reg_t size_reg;
		if (!v_getreg(&size_reg, V_I, V_TEMP))
		    v_fatal("gen field convert out of registers \n");
		v_seti(size_reg, total_size);
		gen_memcpy(src_addr, src_spec->offset, dest_addr,
			   conv->conversions[i].dest_offset, size_reg);
		v_putreg(size_reg, V_I);
	    }
	} else {
	    v_reg_t loop_var;
	    v_label_t loop;

	    int dest_offset = conv->conversions[i].dest_offset;
	    int src_offset = src_spec->offset;
	    IOdata_type dest_type = src_spec->data_type;
	    int dest_size = conv->conversions[i].dest_size;
	    IOgetFieldStruct tmp_spec;
	    tmp_spec = *src_spec;
	    tmp_spec.offset = 0;

	    if (conv->conversions[i].control_field != NULL) {
		/* really a variant array, adjust address like a string */
		tmp_spec.data_type = string_type;
		tmp_spec.size = conv->ioformat->pointer_size;
		dest_type = string_type;
		dest_size = conv->target_pointer_size;
	    }
	    if (dimen1 * dimen2 == 1) {
		gen_subfield_conv(conv, i, tmp_spec, dest_type, assume_align, src_addr, src_offset,
				  dest_size, dest_addr, dest_offset,
				  control_base,
				  final_string_base, src_string_base);
	    } else {
		int src_storage = v_local(V_P);
		int dest_storage = v_local(V_P);
		/* save values of src_addr and dest_addr */
		v_stpi(src_addr, v_lp, src_storage);
		v_stpi(dest_addr, v_lp, dest_storage);

		if ((conv->conversions[i].control_field != NULL) ||
		    (conv->conversions[i].subconversion == NULL)) {
		    if (!v_getreg(&loop_var, V_I, V_TEMP))
			v_fatal("gen field convert out of registers \n");
		} else {
		    /* may call a subconversion in here, use VARs */
		    if (!v_getreg(&loop_var, V_I, V_VAR))
			v_fatal("gen field convert out of registers \n");
		}
		REG_DEBUG(("get %d as loop_var\n", _vrr(loop_var)));
		v_addpi(src_addr, src_addr, src_spec->offset);
		v_addpi(dest_addr, dest_addr,
			conv->conversions[i].dest_offset);

		if (conv->conversions[i].control_field != NULL) {
		    /* 
		     * really a variant array, adjust address like a string 
		     */
		    tmp_spec.data_type = string_type;
		    tmp_spec.size = conv->ioformat->pointer_size;
		}
		/* gen conversion loop */
		loop = v_genlabel();
		v_seti(loop_var, dimen1 * dimen2);
		v_label(loop);
		gen_subfield_conv(conv, i, tmp_spec, dest_type, assume_align,
				  src_addr, 0,
				  dest_size, dest_addr, 0,
				  control_base,
				  final_string_base, src_string_base);

		/* generate end of loop */
		v_subli(loop_var, loop_var, 1);
		v_addpi(src_addr, src_addr, tmp_spec.size);
		v_addpi(dest_addr, dest_addr,
			conv->conversions[i].dest_size);
		if (debug_code_generation) {
		    v_scallv((v_vptr) printf, "%P%p%p",
			     "loopvar = %x, src %x, dest %x\n", loop_var,
			     src_addr, dest_addr);
		}
		v_bgtli(loop_var, 0, loop);
		v_putreg(loop_var, V_I);
		REG_DEBUG(("put %d as loop_var\n", _vrr(loop_var)));

		/* restore values of src_addr and dest_addr */
		v_ldpi(src_addr, v_lp, src_storage);
		v_ldpi(dest_addr, v_lp, dest_storage);
	    }
	}
    }
    return NULL;
}

static void
gen_var_part_conv(conv, i, control_base, assume_align, dest_addr,
		  src_addr, src_string_base, final_string_base)
IOConversionPtr conv;
int i;
int control_base;
int assume_align;
v_reg_t src_addr, dest_addr;
v_reg_t src_string_base, final_string_base;
{
    /* 
     * This section handles converting or moving the body 
     * of a dynamic array or string (the actual pointer was handled 
     * above)
     */
/** handle copying or conversion */
    if (debug_code_generation) {
	v_scallv((v_vptr) printf, "%P%I%p%p%p%p",
		 "varconvpart conversion %d src %lx, dest %lx, src_string_base %lx, final_string_base %lx\n",
	     i, src_addr, dest_addr, src_string_base, final_string_base);
    }
    if (conv->conversions[i].src_field.data_type == string_type) {
	/* generate a call to strcpy to do the move */
	v_label_t end = v_genlabel();
	v_beqpi(src_addr, 0, end);
	v_scalli((v_iptr) strcpy, "%p%p", dest_addr, src_addr);
	v_label(end);
    } else {
	IOFieldPtr src_spec = &conv->conversions[i].src_field;
	/* variant array */
	if (!conv->conversions[i].src_field.byte_swap &&
	    (src_spec->size == conv->conversions[i].dest_size) &&
	    (conv->conversions[i].subconversion == NULL) &&
	    (src_spec->data_type != string_type)) {

	    /* data movement is all that is required */
	    if (conv->conversion_type == copy_strings) {
		v_reg_t size_reg;
		if (!v_getreg(&size_reg, V_I, V_TEMP))
		    v_fatal("gen var convert size out of registers \n");
		/* load control (array size) value */
		v_ldii(size_reg, v_lp, control_base + i * sizeof(int));
		/* scale it by destination size */
		v_mulii(size_reg, (v_reg_type)size_reg,
			conv->conversions[i].dest_size);
		/* generate a memcpy */
		gen_memcpy(src_addr, 0, dest_addr, 0, size_reg);
		v_putreg(size_reg, V_I);
	    } else {
		/* 
		 * not even that is necessary, variant part is
		 * already where it needs to go 
		 */
	    }
	} else {
	    /* must do conversions one by one */
	    int src_storage = v_local(V_P);
	    int dest_storage = v_local(V_P);
	    v_reg_t loop_var;
	    v_reg_t tmp;
	    int dest_size = conv->conversions[i].dest_size;
	    v_label_t loop = v_genlabel();
	    IOdata_type dest_type = src_spec->data_type;

	    /* save values of src_addr and dest_addr */
	    v_stpi(src_addr, v_lp, src_storage);
	    v_stpi(dest_addr, v_lp, dest_storage);

	    if ((conv->conversions[i].control_field != NULL) ||
		(conv->conversions[i].subconversion == NULL)) {
		if (!v_getreg(&loop_var, V_I, V_TEMP))
		    v_fatal("gen var convert loop out of registers \n");
	    } else {
		/* may call a subconversion in here, use VARs */
		if (!v_getreg(&loop_var, V_I, V_VAR))
		    v_fatal("gen var convert loop out of registers \n");
	    }
	    REG_DEBUG(("get %d as var_loop_var\n", _vrr(loop_var)));
	    v_ldii(loop_var, v_lp, control_base + i * sizeof(int));

	    if (!v_getreg(&tmp, V_I, V_TEMP))
		v_fatal("gen var convert tmp out of registers \n");
	    v_mulii(tmp, loop_var, (dest_size - src_spec->size));

	    /* adjust string base !!!!  non-local variation */
	    /* this is because new variant part may change size */
	    v_addi(final_string_base, final_string_base, tmp);
	    v_putreg(tmp, V_I);

	    v_label(loop);
	    if (debug_code_generation) {
		v_scallv((v_vptr) printf, "%P%i%p%p",
			 "varloopvar = %x, src %x, dest %x\n",
			 loop_var, src_addr, dest_addr);
	    }
	    v_savep(src_addr);
	    v_savep(dest_addr);
	    v_savep(loop_var);
	    gen_subfield_conv(conv, i, *src_spec, dest_type,
			      assume_align, src_addr, 0, dest_size,
			      dest_addr, 0,
			      control_base, final_string_base,
			      src_string_base);
	    v_restorep(loop_var);
	    v_restorep(dest_addr);
	    v_restorep(src_addr);
	    v_subli(loop_var, loop_var, 1);
	    v_addpi(src_addr, src_addr, src_spec->size);
	    v_addpi(dest_addr, dest_addr,
		    conv->conversions[i].dest_size);
	    v_bgtli(loop_var, 0, loop);
	    v_putreg(loop_var, V_I);
	    REG_DEBUG(("put %d as loop_var\n", _vrr(loop_var)));
	    /* restore values of src_addr and dest_addr */
	    v_ldpi(src_addr, v_lp, src_storage);
	    v_ldpi(dest_addr, v_lp, dest_storage);
	}
    }
}

#endif
