#include "config.h"
#include <fcntl.h>
#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif
#ifdef HAVE_STRING_H
#include <string.h>
#endif
#include <errno.h>
#include <sys/types.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#include <stdlib.h>
#include <time.h>
#ifdef HAVE_SYS_SOCKET_H
#include <sys/socket.h>
#endif
#ifdef HAVE_SYS_SELECT_H
#include <sys/select.h>
#endif
#ifndef HAVE_WINDOWS_H
#include <sys/un.h>
#include <netinet/in.h>
#include <netdb.h>
#else
#include <winsock.h>
#endif
#ifdef HAVE_SYS_UIO_H
#include <sys/uio.h>
#endif
#include "assert.h"
#include "io.h"
#include "io_interface.h"
#include "io_internal.h"
#include "unix_defs.h"

#if !defined(SOCKET_DEFINED) && !defined(HAVE_WINDOWS_H)
extern int socket ARGS((int domain, int type, int protocol));
#endif
#if !defined(CONNECT_DEFINED) && !defined(HAVE_WINDOWS_H)
extern int connect ARGS((int s, struct sockaddr *name, int namelen));
#endif

static char test_byte = 'T';

extern void
establish_server_connection(iofile, do_fallback)
IOFile iofile;
int do_fallback;
{
    int sock;
    int junk_errno;
    char *junk_str;

    if (os_sockets_init_func) {
	 os_sockets_init_func();
    }
    if ((iofile->server_fd == (void*)-1) || 
	(os_server_write_func(iofile->server_fd, &test_byte, 1, &junk_errno,
			      &junk_str) != 1)) {
	/* reestablish connection, name_str is the machine name */
	struct sockaddr_in sock_addr;
	struct hostent *host_addr;
	char *format_server_host = getenv("FORMAT_SERVER_HOST");

	if (format_server_host == NULL) {
	    format_server_host = FORMAT_SERVER_HOST;	/* from configure */
	}
	if ((sock = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
	  return;
	}
	
	sock_addr.sin_family = AF_INET;
		
	if ((host_addr = gethostbyname(format_server_host)) == NULL) {
	    return;
	}
	memcpy(&sock_addr.sin_addr, host_addr->h_addr, host_addr->h_length);
	sock_addr.sin_port = htons(FS_PORT);

	if (connect(sock, (struct sockaddr *) &sock_addr,
		    sizeof sock_addr) < 0) {

	    if (!do_fallback) return;

	    /* fallback */
	    if ((sock = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
	       return;
	    }
	    format_server_host = "marquesas.cc.gatech.edu";
	    sock_addr.sin_family = AF_INET;
	    if ((host_addr = gethostbyname(format_server_host)) == NULL) {
	        return;
	    }
	    memcpy(&sock_addr.sin_addr, host_addr->h_addr, host_addr->h_length);
	    sock_addr.sin_port = htons(FS_PORT);
	    if (connect(sock, (struct sockaddr *) &sock_addr,
			sizeof sock_addr) < 0) {
	        return;
	    }
	}
	
	iofile->server_fd = (void*)sock;
	server_write_header(iofile);
    }
}
	
extern void (*establish_server_connection_ptr)ARGS((IOFile iofile, int do_fallback));

extern
IOContext
create_IOcontext()
{
    IOFile iofile;

    /* fill in function pointer used in io_formats.c */
    establish_server_connection_ptr = establish_server_connection;

    iofile = new_IOFile();
    iofile->status = OpenNoHeader;
    iofile->IOversion = CURRENT_IO_VERSION;
    iofile->is_context = TRUE;
    reset_read_ahead(iofile);
    return (IOContext) iofile;
}

extern
IOContext
create_IOsubcontext(context)
IOContext context;
{
    IOFile iofile;

    /* fill in function pointer used in io_formats.c */
    establish_server_connection_ptr = establish_server_connection;

    iofile = new_IOFile();
    iofile->status = OpenNoHeader;
    iofile->IOversion = CURRENT_IO_VERSION;
    iofile->is_context = TRUE;
    iofile->master_context = context;
    reset_read_ahead(iofile);
    return (IOContext) iofile;
}
